%* xml_validate.sas                                                               *;
%*                                                                                *;
%* Validate an XML file against a W3C schema file.                                *;
%*                                                                                *;
%* Notes:                                                                         *;
%*  (1) This macro uses PROC GROOVY, and will only run in SAS versions 9.3 or     *;
%*      later.                                                                    *;
%*                                                                                *;
%* @param xsdFile - required - The full path to the W3C XML Schema XML file.      *;
%* @param xmlFolder - required - The full path to the folder with XML files to    *;
%*            validate against the W3C XML Schema.                                *;
%* @param whereClause - optional - A valid SAS WHERE clause used in a datastep to *;
%*            restrict the XML files in the folder to be validated.               *;
%*            Examples:                                                           *;
%*              whereClause=%nrstr(where index(upcase(xmlfile), "DEFINE") eq 0)   *;
%*              whereClause=%nrstr(where index(upcase(xmlfile), "DEFINE") ge 1)   *;
%*                                                                                *;
%* @since 1.7                                                                     *;
%* @exposure external                                                             *;

%macro xml_validate(
  xsdFile=,
  xmlFolder=,
  whereClause=
  ) / des="Validate an XML file against XML Schema";
  
  %*****************************************************************************;
  %*  Parameter checking                                                       *;
  %*****************************************************************************;

  %if %eval(&sysver LT 9.3) %then 
  %do;
    %put NOTE: [CSTLOG%str(MESSAGE).&sysmacroname] PROC GROOVY is only available on SAS 9.3 and higher.;
    %goto exit_error;
  %end;

  %if %sysevalf(%superq(xsdFile)=, boolean) or
      %sysevalf(%superq(xmlFolder)=, boolean) %then
  %do;
    %put ERR%STR(OR): [CSTLOG%str(MESSAGE).&sysmacroname] xsdFile and xmlFolder parameter values are required.;
    %goto exit_error;
  %end;

  %if not %sysfunc(fileexist(&xsdFile)) %then
  %do;
    %* Rule: xsdFile must exist  *;
     %put ERR%STR(OR): [CSTLOG%str(MESSAGE).&sysmacroname] The file xsdFile=&xsdFile does not exist.;
    %goto exit_error;
  %end;

  %let rc = %sysfunc(filename(fileref,&xmlFolder)) ; 
  %if %sysfunc(fexist(&fileref))=0 %then 
  %do;
    %put ERR%STR(OR): [CSTLOG%str(MESSAGE).&sysmacroname] The folder &xmlFolder does not exist.;
    %goto exit_error;
  %end;
  %let rc=%sysfunc(filename(fileref)); 

  %*****************************************************************************;

  %* Get XML files *;
  data xml(keep=xmlfile);
    length xmlfile $400 rc did memcnt 8;
    rc=filename("fref", "&xmlFolder");
    did=dopen("fref");
    memcnt=dnum(did);
    do i = 1 to memcnt;
      xmlfile=cats("&xmlFolder", "/", dread(did,i));
      if scan(upcase(xmlfile), -1, ".") = "XML" then output;
    end;
    rc=dclose(did);
  run;
  
  %* Call Java object *;
  data _null_;
    set xml;
    
    &whereClause;
    
    xsdFile="&xsdFile";
    dcl javaobj j("ValidateXML");
    call j.callVoidMethod("validateXML", xmlFile, xsdFile);
  run;

    
  %exit_error:

%mend xml_validate;
